"""Differentiable Likelihood for CMB Analysis"""

__author__ = ["L. Balkenhol", "C. Trendafilova", "K. Benabed", "S. Galli"]
__version__ = "2.0.3"
__description__ = "Differentiable Likelihood for CMB Analysis."

from .likelihood import *
